import subprocess

import subprocess


def run_uncrustify(file_path: str, config_path: str):
    try:
        # Run uncrustify on the file
        result = subprocess.run(
            [
                "uncrustify",
                "-c",
                config_path,
                "-f",
                file_path,
                "-o",
                file_path,
                "--no-backup",
            ],
            check=True,
        )

        # If the return code is 0, then uncrustify ran successfully
        if result.returncode == 0:
            return
            # print(f"Uncrustify ran successfully on {file_path}.")
        else:
            print(f"Uncrustify failed on {file_path}.")

    except Exception as e:
        print(f"Error running uncrustify: {e}")



from jinja2 import Environment, FileSystemLoader
import yaml
import os
import sys
import stringcase
import shutil
import uncrustify
import clang_format
from pathlib import Path


def generate_export_header_file(
    root_path: str,
    path: str,
    application_name: str,
    export: str,
    export_header_file: str,
    files_to_be_generated: dict[str, bool],
):
    # generate the export header file

    template_env = Environment(loader=FileSystemLoader("templates/repositories"))
    export_header_template = template_env.get_template("export_template.jinja2")

    relative_export_header_file = os.path.join(path, export_header_file)

    if not files_to_be_generated.get(relative_export_header_file, False):
        return

    export_header_file = os.path.join(root_path, relative_export_header_file)

    rendered_template = export_header_template.render(
        application_uppercase_name=stringcase.uppercase(application_name),
        export=export,
        layer_uppercase_name="DOMAIN",
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(export_header_file), exist_ok=True)

    with open(export_header_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {export_header_file}")


def generate_entities_enum_file(
    root_path: str,
    path: str,
    entities_list: list[dict],
    application_cpp_domain_name: str,
    files_to_be_generated: dict[str, bool],
):
    template_env = Environment(loader=FileSystemLoader("templates/entities"))
    entities_enum_template = template_env.get_template(
        "entities_enum_template.h.jinja2"
    )

    relative_entities_enum_file = os.path.join(path, "entities.h")

    if not files_to_be_generated.get(relative_entities_enum_file, False):
        return

    entities_enum_file = os.path.join(root_path, relative_entities_enum_file)

    rendered_template = entities_enum_template.render(
        entities_list=entities_list,
        application_cpp_domain_name=application_cpp_domain_name,
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(entities_enum_file), exist_ok=True)

    with open(entities_enum_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {entities_enum_file}")


def generate_domain_registration_file(
    root_path: str,
    path: str,
    entities_list: list[dict],
    export: str,
    export_header_file: str,
    headers: list[str],
    application_cpp_domain_name: str,
    files_to_be_generated: dict[str, bool],
):
    template_env = Environment(loader=FileSystemLoader("templates/entities"))
    domain_registration_template = template_env.get_template(
        "domain_registration_template.h.jinja2"
    )

    relative_domain_registration_output_file = os.path.join(
        path, "domain_registration.h"
    )
    domain_registration_output_file = os.path.join(
        root_path, relative_domain_registration_output_file
    )

    rendered_template = domain_registration_template.render(
        entities_list=entities_list,
        export=export,
        export_header_file=export_header_file,
        application_cpp_domain_name=application_cpp_domain_name,
        headers=headers,
    )

    if files_to_be_generated.get(relative_domain_registration_output_file, False):
        with open(domain_registration_output_file, "w") as fh:
            fh.write(rendered_template)
            print(f"Successfully wrote file {domain_registration_output_file}")


def generate_cmakelists(
    root_path: str,
    path: str,
    application_name: str,
    export_header_file: str,
    files_to_be_generated: dict[str, bool],
):
    # generate the cmakelists.txt

    template_env = Environment(loader=FileSystemLoader("templates/entities"))
    cmakelists_template = template_env.get_template("cmakelists_template.jinja2")

    relative_cmakelists_file = os.path.join(path, "CMakeLists.txt")

    if not files_to_be_generated.get(relative_cmakelists_file, False):
        return

    cmakelists_file = os.path.join(root_path, relative_cmakelists_file)

    rendered_template = cmakelists_template.render(
        application_spinalcase_name=stringcase.spinalcase(application_name),
        application_uppercase_name=stringcase.uppercase(application_name),
        export_header_file=export_header_file,
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(cmakelists_file), exist_ok=True)

    with open(cmakelists_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {cmakelists_file}")


def generate_entity_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = None,
    uncrustify_config_file: str = None,
):
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    entities_data = manifest_data.get("entities", [])
    entities_list = entities_data.get("list", [])
    # remove entities that are not to be generated
    entities_list = [entity for entity in entities_list]

    global_data = manifest_data.get("global", [])
    application_name = global_data.get("application_name", "")
    application_cpp_domain_name = global_data.get(
        "application_cpp_domain_name", "Undefined"
    )

    export = entities_data.get("export", "")
    export_header_file = entities_data.get("export_header_file", "")
    path = entities_data.get("folder_path", ".")

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    template_env = Environment(loader=FileSystemLoader("templates/entities"))
    template = template_env.get_template("entity_template.h.jinja2")

    # Default initialization values
    default_init_values = {
        "int": "0",
        "float": "0.0f",
        "double": "0.0",
        "bool": "false",
        "QString": "QString()",
        "QDateTime": "QDateTime()",
        "QUuid": "QUuid()",
        "QObject": "nullptr",
        "QList": "QList<>()",
    }

    def isUniqueForeignEntity(field_type: str) -> bool:
        for entity in entities_list:
            name = entity["name"]
            if name == field_type:
                return True

        return False

    def isListOrSetForeignEntity(field_type: str) -> bool:
        if "<" not in field_type:
            return False

        type = field_type.split("<")[1].split(">")[0].strip()

        for entity in entities_list:
            name = entity["name"]
            if name == type:
                return True

        return False

    def determine_relationships(
        entities_by_name: dict[str, dict], entity_name: str
    ) -> list[dict]:
        """
        Determine the relationships of the entity with the other entities. These relationships are defined in the other entities.
        """

        relationships = []

        # Loop through all the entities
        for entity in entities_by_name.values():
            # Loop through all the fields of the entity
            for field in entity["fields"]:
                # If the field is a foreign entity, check if it is the current entity
                if isListOrSetForeignEntity(field["type"]) or isUniqueForeignEntity(
                    field["type"]
                ):
                    if "<" in field["type"]:
                        other_entity_name = (
                            field["type"].split("<")[1].split(">")[0].strip()
                        )
                    else:
                        other_entity_name = field["type"]

                    if (
                        entity_name != other_entity_name
                        and entity_name != entity["name"]
                    ):
                        continue

                    # Determine the cardinality of the relationship
                    cardinality = "ManyUnordered"
                    if isUniqueForeignEntity(field["type"]):
                        cardinality = "One"
                    elif field["type"].startswith("QList") and field.get(
                        "ordered", False
                    ):
                        cardinality = "ManyOrdered"

                    # Determine the relationship type
                    relationship_type = (
                        "OneToOne" if cardinality == "One" else "OneToMany"
                    )

                    # Determine the direction of the relationship
                    direction = "Backward"
                    if entity_name == entity["name"]:
                        direction = "Forward"

                    # Add the relationship to the list
                    relationships.append(
                        {
                            "left_entity_name": entity["name"],
                            "right_entity_name": other_entity_name,
                            "field_name": field["name"],
                            "type": relationship_type,
                            "strength": "Strong"
                            if field.get("strong", True)
                            else "Weak",
                            "cardinality": cardinality,
                            "direction": direction,
                        }
                    )

        return relationships

    generated_files = []
    all_headers = []

    # add "active" field to entities fields in entities_list if entity activable is set to true

    for entity in entities_list:
        if entity.get("activable", False):
            entity["fields"].append(
                {
                    "name": "active",
                    "type": "bool",
                    "need_lazy_loader": False,
                    "name_pascal": "Active",
                    "is_primary_key": False,
                    "is_linked_to_another_entity": False,
                }
            )

    for entity in entities_list:
        name = entity["name"]
        fields = entity["fields"]
        parent = entity.get("parent", "EntityBase")
        relationships = []

        # add other informations to fields
        for field in fields:
            field["name_pascal"] = stringcase.pascalcase(field["name"])
            field["is_primary_key"] = False

            # add need_lazy_loader to fields
            if isListOrSetForeignEntity(field["type"]) or isUniqueForeignEntity(
                field["type"]
            ):
                field["need_lazy_loader"] = True
                field["is_linked_to_another_entity"] = True
            else:
                field["need_lazy_loader"] = False
                field["is_linked_to_another_entity"] = False

        # Get the headers to include based on the field types
        headers = []
        for field in fields:
            field_type = field["type"]
            if isListOrSetForeignEntity(field_type):
                include_header_file = field_type.split("<")[1].split(">")[0].strip()
                header_file = f'"{stringcase.snakecase(include_header_file)}.h"'
                headers.append(header_file)
            elif isUniqueForeignEntity(field_type):
                header_file = f'"{stringcase.snakecase(field_type)}.h"'
                headers.append(header_file)
            elif field_type in ["QString", "QDateTime", "QUuid"]:
                headers.append(f"<{field_type}>")

        # remove duplicates
        headers = list(set(headers))

        parent_header = f'"{stringcase.snakecase(parent)}.h"'
        if parent == "EntityBase":
            parent_header = '"qleany/domain/entity_base.h"'

        # If the parent is a custom entity defined in the manifest, include its fields as well. Also, it will add the fields parent of the parent recursively, butonly if the parent is defined in the
        # manifest. If the parent is not defined in the manifest, it will only add the fields of the current entity. But it adds the field "id" of type "int" from EntityBase.
        # This is done to avoid having to add the field "id" in every entity defined in the manifest.

        # Initialize an empty list for parent fields
        parent_fields = []

        # Check if the parent of the current entity is a custom entity defined in the manifest
        if parent in entities_by_name:
            parent_fields.append(
                {
                    "name": "id",
                    "type": "int",
                    "need_lazy_loader": False,
                    "name_pascal": "Id",
                    "is_primary_key": True,
                    "is_linked_to_another_entity": False,
                }
            )
            # If it is, then get the fields of the parent entity
            parent_fields += entities_by_name[parent]["fields"]

            # If parent has its own parent, recursively add its fields
            while (
                "parent" in entities_by_name[parent]
                and entities_by_name[parent]["parent"] in entities_by_name
            ):
                parent = entities_by_name[parent]["parent"]

                parent_fields += entities_by_name[parent]["fields"]
        else:
            parent_fields.append(
                {
                    "name": "id",
                    "type": "int",
                    "need_lazy_loader": False,
                    "name_pascal": "Id",
                    "is_primary_key": True,
                    "is_linked_to_another_entity": False,
                }
            )

        # add other informations to parent_fields
        for field in parent_fields:
            field["name_pascal"] = stringcase.pascalcase(field["name"])
            field["is_primary_key"] = True if field["name"] == "id" else False

            # add need_lazy_loader to fields
            if isListOrSetForeignEntity(field["type"]) or isUniqueForeignEntity(
                field["type"]
            ):
                field["need_lazy_loader"] = True
                field["is_linked_to_another_entity"] = True
            else:
                field["need_lazy_loader"] = False
                field["is_linked_to_another_entity"] = False

        # only keep the fields from the current entity for initialization
        fields_init_values = ", ".join(
            [
                f"m_{field['name']}({default_init_values.get(field['type'].split('<')[0], '{}')})"
                for field in fields
                if not field["need_lazy_loader"]
            ]
        )
        fields_init_values = (
            ", " + fields_init_values if fields_init_values else fields_init_values
        )

        # use parent fields to initialize parent class in constructor
        parent_init_values = ", ".join(
            [
                f"{field['name']}({default_init_values.get(field['type'].split('<')[0], '{}')})"
                for field in parent_fields
            ]
        )

        # relationships
        relationships = determine_relationships(entities_by_name, name)

        rendered_template = template.render(
            name=name,
            parent=parent,
            fields=fields,
            parent_fields=parent_fields,
            headers=headers,
            parent_header=parent_header,
            export=export,
            export_header_file=export_header_file,
            fields_init_values=fields_init_values,
            parent_init_values=parent_init_values,
            relationships=relationships,
            application_cpp_domain_name=application_cpp_domain_name,
        )

        relative_output_file = os.path.join(path, f"{stringcase.snakecase(name)}.h")
        output_file = os.path.join(root_path, relative_output_file)

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(output_file), exist_ok=True)

        if not files_to_be_generated.get(relative_output_file, False):
            continue

        # Add the generated file to the list
        generated_files.append(output_file)

        with open(output_file, "w") as fh:
            fh.write(rendered_template)
            print(f"Successfully wrote file {output_file}")

        # if uncrustify_config_file:
        #     uncrustify.run_uncrustify(output_file, uncrustify_config_file)
        clang_format.run_clang_format(output_file)

    # generate entities cmake file

    entities_cmake_template = template_env.get_template("entities.cmake.jinja2")

    relative_cmake_output_file = os.path.join(path, "entities.cmake")
    cmake_output_file = os.path.join(root_path, relative_cmake_output_file)

    entities = []
    for entity in entities_list:
        entity_header = f'{stringcase.snakecase(entity["name"])}.h'
        relative_path = os.path.relpath(
            os.path.join(root_path, path, entity_header),
            os.path.dirname(cmake_output_file),
        )
        entities.append(relative_path.replace("\\", "/"))

    rendered_template = entities_cmake_template.render(
        entities=entities,
    )

    if files_to_be_generated.get(relative_cmake_output_file, False):
        with open(cmake_output_file, "w") as fh:
            fh.write(rendered_template)
            print(f"Successfully wrote file {output_file}")

    # generate entities enum file
    generate_entities_enum_file(
        root_path,
        path,
        entities_list,
        application_cpp_domain_name,
        files_to_be_generated,
    )

    # generate domain registration file
    for entity in entities_list:
        name = entity["name"]
        header_file = f'"{stringcase.snakecase(name)}.h"'
        all_headers.append(header_file)

    generate_domain_registration_file(
        root_path,
        path,
        entities_list,
        export,
        export_header_file,
        all_headers,
        application_cpp_domain_name,
        files_to_be_generated,
    )

    generate_cmakelists(
        root_path,
        path,
        manifest_data["global"]["application_name"],
        export_header_file,
        files_to_be_generated,
    )

    # write the export header file

    generate_export_header_file(
        root_path,
        path,
        application_name,
        export,
        export_header_file,
        files_to_be_generated,
    )


def get_files_to_be_generated(
    manifest_file: str, files_to_be_generated: dict[str, bool] = None
) -> list[str]:
    """
    Get the list of files that need to be generated based on the manifest file
    """
    # Read the manifest file
    with open(manifest_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    folder_path = manifest["entities"]["folder_path"]
    export_header_file = manifest["entities"]["export_header_file"]

    # Get the list of files to be generated
    files = []
    for entity in manifest["entities"]["list"]:
        entity_name = entity["name"]
        files.append(
            os.path.join(folder_path, f"{stringcase.snakecase(entity_name)}.h")
        )

    # add list cmake file:
    list_file = os.path.join(folder_path, "entities.cmake")
    files.append(list_file)

    files.append(os.path.join(folder_path, "entities.h"))
    files.append(os.path.join(folder_path, "domain_registration.h"))
    files.append(os.path.join(folder_path, export_header_file))
    files.append(os.path.join(folder_path, "CMakeLists.txt"))

    # strip from files if the value in files_to_be_generated is False
    if files_to_be_generated:
        for path, generate in files_to_be_generated.items():
            if not generate:
                files.remove(path)

    return files


# generate the files into the preview folder
def preview_entity_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = None,
    uncrustify_config_file: str = None,
):
    manifest_preview_file = "temp/manifest_preview.yaml"

    # make a copy of the manifest file into temp/manifest_preview.yaml
    shutil.copy(manifest_file, manifest_preview_file)

    # modify the manifest file to generate the files into the preview folder
    with open(manifest_preview_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    # remove .. from the path
    manifest["entities"]["folder_path"] = manifest["entities"]["folder_path"].replace(
        "..", ""
    )

    # write the modified manifest file
    with open(manifest_preview_file, "w") as fh:
        yaml.dump(manifest, fh)

    root_path = os.path.join(root_path, "qleany_preview")

    # remove .. from the path
    if files_to_be_generated:
        preview_files_to_be_generated = {}
        for path, value in files_to_be_generated.items():
            preview_files_to_be_generated[path.replace("..", "")] = value

        generate_entity_files(
            root_path,
            manifest_preview_file,
            preview_files_to_be_generated,
            uncrustify_config_file,
        )
    else:
        generate_entity_files(
            root_path, manifest_preview_file, {}, uncrustify_config_file
        )


# Main execution
if __name__ == "__main__":
    full_path = Path(__file__).resolve().parent

    # Add the current directory to the path so that we can import the generated files
    sys.path.append(full_path)

    # Set the current directory to the generator directory
    os.chdir(full_path)

    if len(sys.argv) > 1:
        manifest_arg = sys.argv[1]
        if manifest_arg.endswith(".yaml") or manifest_arg.endswith(".yml"):
            manifest_file = manifest_arg
            root_path = Path(manifest_file).parent

            if len(sys.argv) > 2 and sys.argv[2] == "--preview":
                preview_entity_files(root_path, manifest_file)
            else:
                # generate the files
                generate_entity_files(root_path, manifest_file)
        else:
            print("Error: Manifest file must be named 'qleany.yaml' or 'qleany.yml'")
    else:
        print("Error: Please provide the manifest file as an argument")



from jinja2 import Environment, FileSystemLoader
import yaml
import os
import sys
import stringcase
import shutil
import uncrustify
import clang_format
from pathlib import Path


def get_cqrs_dict(
    cqrs_common_cmake_folder_path,
    application_cpp_domain_name,
    feature_by_name,
    entities_by_name,
) -> dict:
    # example of wanted dict
    # generation_dict = {
    #     "features": {
    #         "MyFeature": {
    #             "queries": {
    #                 "MyQuery": {
    #                     "file": os.path.join(
    #                         cqrs_common_cmake_folder_path,
    #                         "MyFeature",
    #                         "queries",
    #                         "my_query.h",
    #                     ),
    #                     "dto_pascal": "MyQueryDTO",
    #                     "dto_snake": "my_query_dto",
    #                 },
    #             },
    #             "commands": {
    #                 "MyCommand": {
    #                     "file": os.path.join(
    #                         cqrs_common_cmake_folder_path,
    #                         "MyFeature",
    #                         "commands",
    #                         "my_command.h",
    #                     ),
    #                     "dto_pascal": "MyCommandDTO",
    #                     "dto_snake": "my_command_dto",
    #                 },
    #             },
    #             "validators": {
    #                 "MyValidator": {
    #                     "file": os.path.join(
    #                         cqrs_common_cmake_folder_path,
    #                         "MyFeature",
    #                         "validators",
    #                         "my_validator.h",
    #                     ),
    #                     "repository_entity_pascal": "MyEntity",
    #                     "repository_entity_snake": "my_entity",
    #                     "repository_entity_camel": "myEntity",
    #                     "validate_id_existence": True,
    #                 }
    #             },
    #             "cmake_folder_path": os.path.join(cqrs_common_cmake_folder_path, "MyFeature"),
    #         },
    #     },
    #     "common_cmake_folder_path": cqrs_common_cmake_folder_path,
    # }

    generation_dict = {
        "features": {},
        "common_cmake_folder_path": cqrs_common_cmake_folder_path,
    }

    for feature_name, feature in feature_by_name.items():
        feature_dict = {
            "queries": {},
            "commands": {},
            "validators": {},
            "cmake_folder_path": os.path.join(
                cqrs_common_cmake_folder_path,
                stringcase.snakecase(feature_name) + "_cqrs",
            ),
        }
        generation_dict["features"][feature_name] = feature_dict

        feature_name_snake = stringcase.snakecase(feature_name)
        generation_dict["features"][feature_name][
            "feature_name_snake"
        ] = feature_name_snake
        feature_name_pascal = stringcase.pascalcase(feature_name)
        generation_dict["features"][feature_name][
            "feature_name_pascal"
        ] = feature_name_pascal
        generation_dict["features"][feature_name][
            "feature_name_camel"
        ] = stringcase.camelcase(feature_name)
        generation_dict["features"][feature_name][
            "feature_name_spinal"
        ] = stringcase.spinalcase(feature_name)

        # add CRUD queries & commands:
        if feature.get("CRUD", False) and feature.get("CRUD", {}).get("enabled", False):
            entity_mappable_with = feature.get("CRUD", {}).get(
                "entity_mappable_with", feature_name_pascal
            )
            entity_mappable_with_snake = stringcase.snakecase(entity_mappable_with)
            entity_mappable_with_pascal = stringcase.pascalcase(entity_mappable_with)
            # get and get_with_details
            if feature.get("CRUD", {}).get("get", {}).get("enabled", False):
                generation_dict["features"][feature_name]["queries"][
                    "Get" + entity_mappable_with_pascal
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "queries",
                        "get_" + entity_mappable_with_snake + "_query.h",
                    ),
                    "name": "Get" + entity_mappable_with_pascal + "Query",
                    "dto_pascal": f"{entity_mappable_with}DTO",
                    "dto_snake": entity_mappable_with_snake + "_dto",
                    "request_is_id": True,
                    "has_dto_in": False,
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

            # add CRUD commands:
            # create
            if feature.get("CRUD", {}).get("create", {}).get("enabled", False):
                generation_dict["features"][feature_name]["commands"][
                    "Create" + entity_mappable_with_pascal
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "commands",
                        "create_" + entity_mappable_with_snake + "_command.h",
                    ),
                    "name": "Create" + entity_mappable_with_pascal + "Command",
                    "dto_pascal": "Create" + entity_mappable_with + "DTO",
                    "dto_snake": "create_" + entity_mappable_with_snake + "_dto",
                    "request_is_id": False,
                    "has_dto_in": True,
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

                # add validator
                generation_dict["features"][feature_name]["validators"][
                    "Create" + entity_mappable_with_pascal
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "validators",
                        "create_" + entity_mappable_with_snake + "_command_validator.h",
                    ),
                    "name": "Create" + entity_mappable_with_pascal + "CommandValidator",
                    "dto_pascal": "Create" + entity_mappable_with + "DTO",
                    "dto_snake": "create_" + entity_mappable_with_snake + "_dto",
                    "request_is_id": False,
                    "has_dto_in": True,
                    "validate_id_existence": False,
                    "validate_id_absence": False,
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                    "repositories": [
                        {
                            "pascal_name": entity_mappable_with,
                            "snake_name": entity_mappable_with_snake,
                            "camel_name": stringcase.camelcase(entity_mappable_with),
                        }
                    ],
                }

            # update
            if feature.get("CRUD", {}).get("update", {}).get("enabled", False):
                generation_dict["features"][feature_name]["commands"][
                    "Update" + entity_mappable_with_pascal
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "commands",
                        "update_" + entity_mappable_with_snake + "_command.h",
                    ),
                    "name": "Update" + entity_mappable_with_pascal + "Command",
                    "dto_pascal": "Update" + entity_mappable_with + "DTO",
                    "dto_snake": "update_" + entity_mappable_with_snake + "_dto",
                    "request_is_id": False,
                    "has_dto_in": True,
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

                # add validator
                generation_dict["features"][feature_name]["validators"][
                    "Update" + entity_mappable_with_pascal
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "validators",
                        "update_" + entity_mappable_with_snake + "_command_validator.h",
                    ),
                    "name": "Update" + entity_mappable_with_pascal + "CommandValidator",
                    "dto_pascal": "Update" + entity_mappable_with + "DTO",
                    "dto_snake": "update_" + entity_mappable_with_snake + "_dto",
                    "request_is_id": False,
                    "has_dto_in": True,
                    "validate_id_existence": True,
                    "validate_id_absence": False,
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                    "repositories": [
                        {
                            "pascal_name": entity_mappable_with,
                            "snake_name": entity_mappable_with_snake,
                            "camel_name": stringcase.camelcase(entity_mappable_with),
                        }
                    ],
                }

            # remove
            if feature.get("CRUD", {}).get("remove", {}).get("enabled", False):
                generation_dict["features"][feature_name]["commands"][
                    "Remove" + entity_mappable_with_pascal
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "commands",
                        "remove_" + entity_mappable_with_snake + "_command.h",
                    ),
                    "name": "Remove" + entity_mappable_with_pascal + "Command",
                    "dto_pascal": "Remove" + entity_mappable_with + "DTO",
                    "dto_snake": "remove_" + entity_mappable_with_snake + "_dto",
                    "request_is_id": True,
                    "has_dto_in": False,
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

                # add validator
                generation_dict["features"][feature_name]["validators"][
                    "Remove" + entity_mappable_with_pascal
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "validators",
                        "remove_" + entity_mappable_with_snake + "_command_validator.h",
                    ),
                    "name": "Remove" + entity_mappable_with + "CommandValidator",
                    "dto_pascal": "Remove" + entity_mappable_with + "DTO",
                    "dto_snake": "remove_" + entity_mappable_with_snake + "_dto",
                    "request_is_id": True,
                    "has_dto_in": False,
                    "validate_id_existence": True,
                    "validate_id_absence": False,
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                    "repositories": [
                        {
                            "pascal_name": entity_mappable_with,
                            "snake_name": entity_mappable_with_snake,
                            "camel_name": stringcase.camelcase(entity_mappable_with),
                        }
                    ],
                }

        # add custom queries & commands
        custom_queries = feature.get("queries", [])
        for custom_query in custom_queries:
            custom_query_name = custom_query.get("name", "")
            custom_query_snake_name = stringcase.snakecase(custom_query_name)
            custom_query_pascal_name = stringcase.pascalcase(custom_query_name)
            dto_in_type_prefix = (
                custom_query.get("dto", {}).get("in", {}).get("type_prefix", "")
            )
            if custom_query_name:
                generation_dict["features"][feature_name]["queries"][
                    custom_query_name
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "queries",
                        custom_query_snake_name + "_query.h",
                    ),
                    "name": custom_query_pascal_name + "Query",
                    "dto_pascal": stringcase.pascalcase(dto_in_type_prefix + "DTO"),
                    "dto_snake": stringcase.snakecase(dto_in_type_prefix + "_dto"),
                    "request_is_id": custom_query.get("request_is_id", False),
                    "has_dto_in": custom_query.get("dto", {})
                    .get("in", {})
                    .get("enabled", True),
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

            # if validator is enabled
            if custom_query.get("validator", {}).get("enabled", False):
                custom_queries_entities = custom_query.get("entities", [])
                repositories = [
                    {
                        "camel_name": stringcase.camelcase(x),
                        "pascal_name": stringcase.pascalcase(x),
                        "snake_name": stringcase.snakecase(x),
                    }
                    for x in custom_queries_entities
                ]
                generation_dict["features"][feature_name]["validators"][
                    custom_query_name
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "validators",
                        custom_query_snake_name + "_query_validator.h",
                    ),
                    "name": custom_query_pascal_name + "QueryValidator",
                    "dto_pascal": stringcase.pascalcase(dto_in_type_prefix + "DTO"),
                    "dto_snake": stringcase.snakecase(dto_in_type_prefix + "_dto"),
                    "request_is_id": custom_query.get("request_is_id", False),
                    "has_dto_in": custom_query.get("dto", {})
                    .get("in", {})
                    .get("enabled", True),
                    "validate_id_existence": custom_query.get("validator", {}).get(
                        "validate_id_existence", False
                    ),
                    "validate_id_absence": custom_query.get("validator", {}).get(
                        "validate_id_absence", False
                    ),
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "repositories": repositories,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

        custom_commands = feature.get("commands", [])
        for custom_command in custom_commands:
            custom_command_name = custom_command.get("name", "")
            custom_command_snake_name = stringcase.snakecase(custom_command_name)
            custom_command_pascal_name = stringcase.pascalcase(custom_command_name)
            dto_in_type_prefix = (
                custom_command.get("dto", {}).get("in", {}).get("type_prefix", "")
            )
            if custom_command_name:
                generation_dict["features"][feature_name]["commands"][
                    custom_command_name
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "commands",
                        custom_command_snake_name + "_command.h",
                    ),
                    "name": custom_command_pascal_name + "Command",
                    "dto_pascal": stringcase.pascalcase(dto_in_type_prefix + "DTO"),
                    "dto_snake": stringcase.snakecase(dto_in_type_prefix + "_dto"),
                    "request_is_id": custom_command.get("request_is_id", False),
                    "has_dto_in": custom_command.get("dto", {})
                    .get("in", {})
                    .get("enabled", True),
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

            # if validator is enabled
            if custom_command.get("validator", {}).get("enabled", False):
                custom_command_entities = custom_command.get("entities", [])
                repositories = [
                    {
                        "camel_name": stringcase.camelcase(x),
                        "pascal_name": stringcase.pascalcase(x),
                        "snake_name": stringcase.snakecase(x),
                    }
                    for x in custom_command_entities
                ]
                generation_dict["features"][feature_name]["validators"][
                    custom_command_name
                ] = {
                    "file": os.path.join(
                        cqrs_common_cmake_folder_path,
                        feature_name_snake + "_cqrs",
                        feature_name_snake,
                        "validators",
                        custom_command_snake_name + "_command_validator.h",
                    ),
                    "name": custom_command_pascal_name + "CommandValidator",
                    "dto_pascal": stringcase.pascalcase(dto_in_type_prefix + "DTO"),
                    "dto_snake": stringcase.snakecase(dto_in_type_prefix + "_dto"),
                    "request_is_id": custom_command.get("request_is_id", False),
                    "has_dto_in": custom_command.get("dto", {})
                    .get("in", {})
                    .get("enabled", True),
                    "validate_id_existence": custom_command.get("validator", {}).get(
                        "validate_id_existence", False
                    ),
                    "validate_id_absence": custom_command.get("validator", {}).get(
                        "validate_id_absence", False
                    ),
                    "feature_name_snake": feature_name_snake,
                    "feature_name_pascal": feature_name_pascal,
                    "repositories": repositories,
                    "application_cpp_domain_name": application_cpp_domain_name,
                }

        # add all files in each generation_dict[feature]["files"]
        generation_dict["features"][feature_name]["files"] = []
        for query_name, query in generation_dict["features"][feature_name][
            "queries"
        ].items():
            generation_dict["features"][feature_name]["files"].append(query["file"])
        for command_name, command in generation_dict["features"][feature_name][
            "commands"
        ].items():
            generation_dict["features"][feature_name]["files"].append(command["file"])
        for validator_name, validator in generation_dict["features"][feature_name][
            "validators"
        ].items():
            generation_dict["features"][feature_name]["files"].append(validator["file"])
        # add the feature cmake file
        generation_dict["features"][feature_name]["files"].append(
            os.path.join(
                generation_dict["features"][feature_name]["cmake_folder_path"],
                "CMakeLists.txt",
            )
        )

    # add all files into the generation dict["files"]
    generation_dict["files"] = []
    for feature_name, feature in generation_dict["features"].items():
        for file in feature["files"]:
            generation_dict["files"].append(file)

    return generation_dict


def generate_query(
    root_path: str, query: dict, files_to_be_generated: dict[str, bool] = None
):
    template_env = Environment(loader=FileSystemLoader("templates/CQRS"))
    template = template_env.get_template("query_template.jinja2")

    file_path = query["file"]

    if not files_to_be_generated.get(file_path, False):
        return file_path

    file_path = os.path.join(root_path, file_path)

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    with open(file_path, "w") as f:
        f.write(template.render(query=query))
        print(f"Successfully wrote file {file_path}")

    return file_path


def generate_command(
    root_path: str, command: dict, files_to_be_generated: dict[str, bool] = None
):
    template_env = Environment(loader=FileSystemLoader("templates/CQRS"))
    template = template_env.get_template("command_template.jinja2")

    file_path = command["file"]

    if not files_to_be_generated.get(file_path, False):
        return file_path

    file_path = os.path.join(root_path, file_path)

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    with open(file_path, "w") as f:
        f.write(template.render(command=command))
        print(f"Successfully wrote file {file_path}")

    return file_path


def generate_validator(
    root_path: str, validator: dict, files_to_be_generated: dict[str, bool] = None
):
    template_env = Environment(loader=FileSystemLoader("templates/CQRS"))
    template = template_env.get_template("validator_template.jinja2")

    file_path = validator["file"]

    if not files_to_be_generated.get(file_path, False):
        return file_path

    file_path = os.path.join(root_path, file_path)

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    with open(file_path, "w") as f:
        f.write(template.render(validator=validator))
        print(f"Successfully wrote file {file_path}")

    return file_path


def generate_feature_cmake(
    root_path: str,
    application_name: str,
    feature: dict,
    files_to_be_generated: dict[str, bool] = None,
):
    template_env = Environment(loader=FileSystemLoader("templates/CQRS"))
    template = template_env.get_template("cmakelists_template.jinja2")

    cmake_file_path = os.path.join(feature["cmake_folder_path"], "CMakeLists.txt")
    files = feature["files"]

    if not files_to_be_generated.get(cmake_file_path, False):
        return cmake_file_path

    # strip the cmakelists file from the files list
    files = [file for file in files if file != cmake_file_path]

    cmake_file_path = os.path.join(root_path, cmake_file_path)

    ## Convert the file path to be relative to the directory of the cmakelists
    relative_generated_files = []
    for file_path in files:
        relative_generated_file = os.path.relpath(
            os.path.join(root_path, file_path), os.path.dirname(cmake_file_path)
        )
        relative_generated_files.append(relative_generated_file.replace("\\", "/"))

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(cmake_file_path), exist_ok=True)

    with open(cmake_file_path, "w") as f:
        f.write(
            template.render(
                feature=feature,
                files=relative_generated_files,
                application_name=application_name,
            )
        )
        print(f"Successfully wrote file {cmake_file_path}")

    return cmake_file_path


def generate_cmake(
    root_path: str, cqrs_dict: dict, files_to_be_generated: dict[str, bool] = None
):
    # generate common cmakelists.txt
    common_cmake_folder_path = cqrs_dict["common_cmake_folder_path"]

    relative_common_cmakelists_file = os.path.join(
        common_cmake_folder_path, "CMakeLists.txt"
    )
    common_cmakelists_file = os.path.join(root_path, relative_common_cmakelists_file)

    if files_to_be_generated.get(relative_common_cmakelists_file, False):
        ## After the loop, write the list of features folders to the common cmakelists.txt
        with open(common_cmakelists_file, "w") as fh:
            for feature_name, _ in cqrs_dict["features"].items():
                fh.write(
                    f"add_subdirectory({stringcase.snakecase(feature_name)}_cqrs)"
                    + "\n"
                )
            print(f"Successfully wrote file {common_cmakelists_file}")

    return common_cmakelists_file


def generate_cqrs_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = None,
    uncrustify_config_file: str = None,
):
    """
    Get the list of files that need to be generated based on the manifest file
    """
    # Read the manifest file
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    cqrs_data = manifest_data.get("CQRS", [])
    cqrs_common_cmake_folder_path = cqrs_data.get("common_cmake_folder_path", "")

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )
    application_name = stringcase.spinalcase(application_name)
    application_cpp_domain_name = manifest_data.get("global", {}).get(
        "application_cpp_domain_name", "Undefined"
    )

    application_data = manifest_data.get("application", [])
    feature_list = application_data.get("features", [])

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    entities_data = manifest_data.get("entities", [])
    entities_list = entities_data.get("list", [])

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    cqrs_dict = get_cqrs_dict(
        cqrs_common_cmake_folder_path,
        application_cpp_domain_name,
        feature_by_name,
        entities_by_name,
    )

    for feature_name, feature in cqrs_dict["features"].items():
        generated_files = []
        for query_name, query in feature["queries"].items():
            generated_files.append(
                generate_query(root_path, query, files_to_be_generated)
            )
        for command_name, command in feature["commands"].items():
            generated_files.append(
                generate_command(root_path, command, files_to_be_generated)
            )
        for validator_name, validator in feature["validators"].items():
            generated_files.append(
                generate_validator(root_path, validator, files_to_be_generated)
            )

        for file in generated_files:
            # if uncrustify_config_file and files_to_be_generated.get(file, False):
            #     uncrustify.run_uncrustify(file, uncrustify_config_file)
            if files_to_be_generated.get(file, False):
                clang_format.run_clang_format(file)

        # generate feature CMakeLists.txt
        generate_feature_cmake(
            root_path, application_name, feature, files_to_be_generated
        )

    # generate CMakeLists.txt
    generate_cmake(root_path, cqrs_dict, files_to_be_generated)


def get_files_to_be_generated(
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = None,
    uncrustify_config_file: str = None,
) -> list[str]:
    # Read the manifest file
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    cqrs_data = manifest_data.get("CQRS", [])
    cqrs_common_cmake_folder_path = cqrs_data.get("common_cmake_folder_path", "")

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )
    application_name = stringcase.spinalcase(application_name)
    application_cpp_domain_name = manifest_data.get("global", {}).get(
        "application_cpp_domain_name", "Undefined"
    )

    application_data = manifest_data.get("application", [])
    feature_list = application_data.get("features", [])

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    entities_data = manifest_data.get("entities", [])
    entities_list = entities_data.get("list", [])

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    cqrs_dict = get_cqrs_dict(
        cqrs_common_cmake_folder_path,
        application_cpp_domain_name,
        feature_by_name,
        entities_by_name,
    )

    files = cqrs_dict["files"]

    # generate common cmakelists.txt
    dto_cmakelists_file = os.path.join(cqrs_common_cmake_folder_path, "CMakeLists.txt")

    files.append(dto_cmakelists_file)

    # strip from files if the value in files_to_be_generated is False
    if files_to_be_generated:
        for path, generate in files_to_be_generated.items():
            if not generate:
                files.remove(path)

    return files


# generate the files into the preview folder
def preview_cqrs_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = None,
    uncrustify_config_file: str = None,
):
    manifest_preview_file = "temp/manifest_preview.yaml"

    # make a copy of the manifest file into temp/manifest_preview.yaml
    shutil.copy(manifest_file, manifest_preview_file)

    # modify the manifest file to generate the files into the preview folder
    with open(manifest_preview_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    # remove .. from the path
    manifest["CQRS"]["common_cmake_folder_path"] = manifest["CQRS"][
        "common_cmake_folder_path"
    ].replace("..", "")
    # write the modified manifest file
    with open(manifest_preview_file, "w") as fh:
        yaml.dump(manifest, fh)

    root_path = os.path.join(root_path, "qleany_preview")

    # remove .. from the path
    if files_to_be_generated:
        preview_files_to_be_generated = {}
        for path, value in files_to_be_generated.items():
            preview_files_to_be_generated[path.replace("..", "")] = value

        generate_cqrs_files(
            root_path,
            manifest_preview_file,
            preview_files_to_be_generated,
            uncrustify_config_file,
        )

    else:
        generate_cqrs_files(
            root_path, manifest_preview_file, {}, uncrustify_config_file
        )


# Main execution
if __name__ == "__main__":
    full_path = Path(__file__).resolve().parent

    # Add the current directory to the path so that we can import the generated files
    sys.path.append(full_path)

    # Set the current directory to the generator directory
    os.chdir(full_path)

    if len(sys.argv) > 1:
        manifest_arg = sys.argv[1]
        if manifest_arg.endswith(".yaml") or manifest_arg.endswith(".yml"):
            manifest_file = manifest_arg
            root_path = Path(manifest_file).parent

            if len(sys.argv) > 2 and sys.argv[2] == "--preview":
                preview_cqrs_files(root_path, manifest_file)
            else:
                # generate the files
                generate_cqrs_files(root_path, manifest_file)
        else:
            print("Error: Manifest file must be named 'qleany.yaml' or 'qleany.yml'")
    else:
        print("Error: Please provide the manifest file as an argument")



import yaml
from PySide6.QtWidgets import (
    QMainWindow,
    QGraphicsView,
    QGraphicsScene,
    QGraphicsRectItem,
    QGraphicsTextItem,
    QGraphicsLineItem,
)
from PySide6.QtCore import Qt, QRectF
from PySide6.QtGui import QPen, QColor, QFont


class EntityRelationshipWindow(QMainWindow):
    def __init__(self, file_path, parent=None):
        super(EntityRelationshipWindow, self).__init__(parent)
        self.setWindowTitle("Entity Relationship Diagram")

        self.view = QGraphicsView(self)
        self.scene = QGraphicsScene()
        self.view.setScene(self.scene)

        self.entity_items = {}
        self.load_yaml(file_path)

        # set the window size to the scene size but not less than 800x600 and not more than 1920x1080
        self.setFixedSize(self.scene.width() + 20, self.scene.height() + 20)
        self.setMinimumSize(800, 600)
        self.setMaximumSize(1920, 1080)

        self.setCentralWidget(self.view)

    def load_yaml(self, file_path):
        with open(file_path, "r") as f:
            data = yaml.load(f, Loader=yaml.FullLoader)

        entities = data["entities"]["list"]

        x, y = 0, 0

        for entity in entities:
            rect_height = 50 + len(entity["fields"]) * 20
            rect = QGraphicsRectItem(QRectF(x, y, 200, rect_height))
            rect.setZValue(1)
            rect.setBrush(QColor("lightgray"))
            rect.setPen(QPen(Qt.black))
            self.scene.addItem(rect)

            title = f"<b>{entity['name']}</b>"
            entity_name = QGraphicsTextItem(rect)
            entity_name.setHtml(title)
            entity_name.setPos(x + 10, y + 10)
            self.scene.addItem(entity_name)

            y_field = y + 30
            for field in entity["fields"]:
                field_text = f"{field['name']} : {field['type']}"
                field_item = QGraphicsTextItem(field_text, rect)
                field_item.setPos(x + 10, y_field)
                field_item.setZValue(2)
                field_item.setDefaultTextColor(QColor("black"))
                field_item.setFont(QFont("Arial", 10))

                self.scene.addItem(field_item)
                y_field += 20

            self.entity_items[entity["name"]] = rect

            y += rect_height + 50

        for entity in entities:
            parent = entity.get("parent")
            if parent and parent != "EntityBase":
                parent_item = self.entity_items[parent]
                child_item = self.entity_items[entity["name"]]
                x1, y1 = parent_item.rect().center().x(), parent_item.rect().bottom()
                x2, y2 = child_item.rect().center().x(), child_item.rect().top()

                line_type = (
                    Qt.SolidLine
                    if any(field.get("strong", False) for field in entity["fields"])
                    else Qt.DotLine
                )
                line = QGraphicsLineItem(x1, y1, x2, y2)
                pen = QPen()
                pen.setStyle(line_type)
                line.setPen(pen)
                line.setZValue(0)
                self.scene.addItem(line)



from jinja2 import Environment, FileSystemLoader
import yaml
import os
import sys
import stringcase
import shutil
import uncrustify
import clang_format
from pathlib import Path


def get_generation_dict(
    real_imports_folder_path: str,
    mock_imports_folder_path: str,
    feature_by_name: dict,
    application_name: str,
) -> dict:
    generation_dict = {}
    generation_dict["controllers"] = {}
    for feature_name, feature_data in feature_by_name.items():
        feature_snake_name = stringcase.snakecase(feature_name)
        feature_pascal_name = stringcase.pascalcase(feature_name)
        feature_camel_name = stringcase.camelcase(feature_name)
        generation_dict["controllers"][feature_pascal_name] = {
            "mock_presenter_file": os.path.join(
                mock_imports_folder_path,
                "Presenter",
                f"{feature_pascal_name}Controller.qml",
            ),
            "mock_template_path": "QML/mock_imports/presenter/",
            "mock_template": "controller.qml.jinja2",
            "real_presenter_file": os.path.join(
                real_imports_folder_path,
                "presenter",
                f"foreign_{feature_snake_name}_controller.h",
            ),
            "real_template_path": "QML/real_imports/presenter/",
            "real_template": "foreign_controller.h.jinja2",
            "feature_pascal_name": feature_pascal_name,
            "feature_camel_name": feature_camel_name,
            "feature_snake_name": feature_snake_name,
            "CRUD": feature_data.get("CRUD", {}),
        }

    # add mock_custom_functions to the generation_dict["controllers"][feature_pascal_name] dict
    for feature_name, feature_data in feature_by_name.items():
        feature_pascal_name = stringcase.pascalcase(feature_name)
        generation_dict["controllers"][feature_pascal_name][
            "mock_custom_functions"
        ] = []
        commands = feature_data.get("commands", [])
        if commands:
            for command in commands:
                generation_dict["controllers"][feature_pascal_name][
                    "mock_custom_functions"
                ] += [stringcase.camelcase(command["name"])]

        queries = feature_data.get("queries", [])
        if queries:
            for query in queries:
                generation_dict["controllers"][feature_pascal_name][
                    "mock_custom_functions"
                ] += [stringcase.camelcase(query["name"])]

    # add qmldir:
    qmldir_file = os.path.join(mock_imports_folder_path, "Presenter", "qmldir")

    generation_dict["qmldir_file"] = qmldir_file

    # add CMakelists.txt:
    common_cmake_file = os.path.join(
        real_imports_folder_path, "presenter", "CMakeLists.txt"
    )
    generation_dict["common_cmake_file"] = common_cmake_file

    # add "mock_presenter_file" and "real_presenter_file" to the generation_dict["real_presenter_files"] list
    generation_dict["real_presenter_files"] = []
    generation_dict["mock_presenter_files"] = []
    for _, controller in generation_dict["controllers"].items():
        generation_dict["real_presenter_files"].append(
            controller["real_presenter_file"]
        )
        generation_dict["mock_presenter_files"].append(
            controller["mock_presenter_file"]
        )

    # add application_name
    generation_dict["application_name"] = application_name

    return generation_dict


def generate_mock_controller_file(
    root_path: str,
    controller: dict,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock controller file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(controller["mock_presenter_file"], False):
        return

    # Create the jinja2 environment
    template_path = os.path.join("templates", controller["mock_template_path"])
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template(controller["mock_template"])

    crud = controller["CRUD"]
    crud_enabled = crud.get("enabled", False)
    get_enabled = crud.get("get", {}).get("enabled", False)
    get_all_enabled = crud.get("get_all", {}).get("enabled", False)
    create_enabled = crud.get("create", {}).get("enabled", False)
    update_enabled = crud.get("update", {}).get("enabled", False)
    remove_enabled = crud.get("remove", {}).get("enabled", False)

    # Render the template
    output = template.render(
        controller=controller,
        crud=crud,
        crud_enabled=crud_enabled,
        get_enabled=get_enabled,
        get_all_enabled=get_all_enabled,
        create_enabled=create_enabled,
        update_enabled=update_enabled,
        remove_enabled=remove_enabled,
    )

    output_file = os.path.join(root_path, controller["mock_presenter_file"])

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)


def generate_mock_qmldir_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    # generate the mock qmldir file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(generation_dict["qmldir_file"], False):
        return

    # Create the jinja2 environment
    env = Environment(loader=FileSystemLoader("templates/QML/mock_imports/presenter/"))
    # Load the template
    template = env.get_template("qmldir_template.jinja2")

    singleton_list = []
    for _, controller in generation_dict["controllers"].items():
        name = controller["feature_pascal_name"]
        singleton_list.append(f"singleton {name}Controller 1.0 {name}Controller.qml")

    # Render the template
    output = template.render(singleton_list=singleton_list)

    output_file = os.path.join(root_path, generation_dict["qmldir_file"])

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)


def generate_real_controller_file(
    root_path: str,
    controller: dict,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
    uncrustify_config_file: str,
):
    # generate the real controller file if in the files_to_be_generated dict the value is True
    real_presenter_file = controller["real_presenter_file"]

    if not files_to_be_generated.get(real_presenter_file, False):
        return

    # Create the jinja2 environment
    template_path = os.path.join("templates", controller["real_template_path"])
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template(controller["real_template"])

    # Render the template
    output = template.render(controller=controller)

    output_file = os.path.join(root_path, real_presenter_file)

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    # if uncrustify_config_file:
    #     uncrustify.run_uncrustify(real_presenter_file, uncrustify_config_file)
    clang_format.run_clang_format(output_file)


def generate_real_cmakelists_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    common_cmake_file = generation_dict["common_cmake_file"]

    # generate the real cmakelists file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(common_cmake_file, False):
        return

    output_file = os.path.join(root_path, common_cmake_file)

    # Create the jinja2 environment
    env = Environment(loader=FileSystemLoader("templates/QML/real_imports/presenter/"))
    # Load the template
    template = env.get_template("cmakelists_template.jinja2")

    files = generation_dict["real_presenter_files"]
    relative_files = []
    for file in files:
        relative_files.append(
            os.path.relpath(
                os.path.join(root_path, file), os.path.dirname(output_file)
            ).replace("\\", "/")
        )

    # Render the template
    output = template.render(
        files=relative_files,
        application_name=generation_dict["application_name"],
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)


def generate_qml_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = {},
    uncrustify_config_file: str = None,
):
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )
    application_name = stringcase.spinalcase(application_name)

    application_data = manifest_data.get("application", [])
    feature_list = application_data.get("features", [])

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    qml_data = manifest_data.get("qml", [])

    generation_dict = get_generation_dict(
        qml_data["real_imports_folder_path"],
        qml_data["mock_imports_folder_path"],
        feature_by_name,
        application_name,
    )

    # generate mock files
    for _, controller in generation_dict["controllers"].items():
        generate_mock_controller_file(
            root_path, controller, generation_dict, files_to_be_generated
        )

    # generate mock qmldir file
    generate_mock_qmldir_file(root_path, generation_dict, files_to_be_generated)

    # generate real files
    for _, controller in generation_dict["controllers"].items():
        generate_real_controller_file(
            root_path,
            controller,
            generation_dict,
            files_to_be_generated,
            uncrustify_config_file,
        )

    # generate real CMakeLists.txt file
    generate_real_cmakelists_file(root_path, generation_dict, files_to_be_generated)


def get_files_to_be_generated(
    manifest_file: str, files_to_be_generated: dict[str, bool] = {}
) -> list[str]:
    """
    Get the list of files that need to be generated based on the manifest file
    """
    # Read the manifest file
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )
    application_name = stringcase.spinalcase(application_name)

    application_data = manifest_data.get("application", [])
    feature_list = application_data.get("features", [])

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    qml_data = manifest_data.get("qml", [])

    files = []
    generation_dict = get_generation_dict(
        qml_data["real_imports_folder_path"],
        qml_data["mock_imports_folder_path"],
        feature_by_name,
        application_name,
    )
    files += generation_dict["real_presenter_files"]
    files += generation_dict["mock_presenter_files"]

    # for _, controller in generation_dict["controllers"].items():
    #     files += feature["real_presenter_files"]

    # # add CMakelists.txt:
    common_cmake_file = generation_dict["common_cmake_file"]
    files.append(common_cmake_file)

    # # add qmldir:
    qmldir_file = generation_dict["qmldir_file"]
    files.append(qmldir_file)

    # strip from files if the value in files_to_be_generated is False
    if files_to_be_generated:
        for path, generate in files_to_be_generated.items():
            if not generate:
                files.remove(path)

    return files


# generate the files into the preview folder
def preview_qml_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = {},
    uncrustify_config_file: str = None,
):
    manifest_preview_file = "temp/manifest_preview.yaml"

    # make a copy of the manifest file into temp/manifest_preview.yaml
    shutil.copy(manifest_file, manifest_preview_file)

    # modify the manifest file to generate the files into the preview folder
    with open(manifest_preview_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    # remove .. from the path
    manifest["qml"]["mock_imports_folder_path"] = manifest["qml"][
        "mock_imports_folder_path"
    ].replace("..", "")
    manifest["qml"]["real_imports_folder_path"] = manifest["qml"][
        "real_imports_folder_path"
    ].replace("..", "")

    # write the modified manifest file
    with open(manifest_preview_file, "w") as fh:
        yaml.dump(manifest, fh)

    root_path = os.path.join(root_path, "qleany_preview")

    # remove .. from the path
    if files_to_be_generated:
        preview_files_to_be_generated = {}
        for path, value in files_to_be_generated.items():
            preview_files_to_be_generated[path.replace("..", "")] = value

        generate_qml_files(
            root_path,
            manifest_preview_file,
            preview_files_to_be_generated,
            uncrustify_config_file,
        )

    else:
        generate_qml_files(root_path, manifest_preview_file, {}, uncrustify_config_file)


# Main execution
if __name__ == "__main__":
    full_path = Path(__file__).resolve().parent

    # Add the current directory to the path so that we can import the generated files
    sys.path.append(full_path)

    # Set the current directory to the generator directory
    os.chdir(full_path)

    if len(sys.argv) > 1:
        manifest_arg = sys.argv[1]
        if manifest_arg.endswith(".yaml") or manifest_arg.endswith(".yml"):
            manifest_file = manifest_arg
            root_path = Path(manifest_file).parent

            if len(sys.argv) > 2 and sys.argv[2] == "--preview":
                preview_qml_files(root_path, manifest_file)
            else:
                generate_qml_files(root_path, manifest_file)
        else:
            print("Error: Manifest file must be named 'qleany.yaml' or 'qleany.yml'")
    else:
        print("Error: Please provide the manifest file as an argument")
